import { Grammar } from "./grammars/Grammar";
import { Connection } from "../Connection";
import { Expression } from "./Expression";
export interface Aggregate {
    Func: string;
    Column: string;
}
export declare enum WhereType {
    Basic = 0,
    In = 1,
    Null = 2,
    Nested = 3,
    Column = 4,
    Raw = 5,
}
export interface Where {
    Type: WhereType;
    Conjunction: string;
    Not?: boolean;
    Column?: string;
    Operator?: string;
    Value?: any;
    Values?: any[];
    Sql?: string;
    Query?: Builder;
}
export interface Join {
    Type: string;
    Table: string;
    Query?: Builder;
}
export interface Order {
    Column?: string;
    Direction?: string;
    Sql?: string;
}
export interface Group {
    Column?: string;
    Sql?: string;
}
export interface Having {
    Conjunction: string;
    Column?: string;
    Operator?: string;
    Value?: any;
    Sql?: string;
}
export interface Bindings {
    Select: any[];
    Join: any[];
    Update: any[];
    Where: any[];
    Having: any[];
    Values: any[];
}
export interface Map {
    [key: string]: any;
}
export declare class Builder {
    connection: Connection;
    _aggregate: Aggregate;
    _distinct: boolean;
    _columns: [string | Expression];
    _from: string;
    _joins: Join[];
    _wheres: Where[];
    _groups: Group[];
    _havings: Having[];
    _orders: Order[];
    _limit: number;
    _offset: number;
    _bindings: Bindings;
    /**
     * All of the available clause operators
     */
    operators: string[];
    protected _lastSql: string;
    /**
     * Grammar
     */
    protected _grammar: Grammar;
    constructor(connection: Connection);
    newQuery(): Builder;
    select(columns?: [string | Expression]): this;
    addSelect(column: string | Expression): this;
    selectRaw(sql: string, bindings?: any[]): this;
    table(table: string): this;
    join(table: string, first: string | ((q: Builder) => void), operator?: string, second?: string, type?: string): this;
    crossJoin(table: string): this;
    leftJoin(table: string, first: string | ((q: Builder) => void), operator?: string, second?: string): this;
    rightJoin(table: string, first: string | ((q: Builder) => void), operator?: string, second?: string): this;
    where(column: string | ((query: Builder) => void), operator?: any, value?: any, conj?: string): this;
    orWhere(column: string | ((query: Builder) => void), operator?: any, value?: any): this;
    whereColumn(first: string, operator: string, second?: string, conj?: string): this;
    on(first: string, operator: string, second?: string, conj?: string): this;
    orOn(first: string, operator: string, second?: string): this;
    orWhereColumn(first: string, operator: string, second?: string, conj?: string): this;
    whereBetween(column: string, values: any[], conj?: string, not?: boolean): this;
    whereRaw(sql: string, bindings?: any[], conj?: string): this;
    orWhereRaw(sql: string, bindings?: any[]): this;
    whereNested(fn: (query: Builder) => void, conj?: string): this;
    orWhereNested(fn: (query: Builder) => void): this;
    whereNull(column: string, conj?: string, not?: boolean): this;
    orWhereNull(column: string, not?: boolean): this;
    whereNotNull(column: string, conj?: string): this;
    orWhereNotNull(column: string): this;
    whereIn(column: string, values: any[], conj?: string, not?: boolean): this;
    orWhereIn(column: string, values: any[], not?: boolean): this;
    whereNotIn(column: string, values: any[], conj?: string): this;
    orWhereNotIn(column: string, values: any[]): this;
    groupBy(column: string): this;
    groupByRaw(sql: string): this;
    having(column: string, operator: any, value?: any, conj?: string): this;
    orHaving(column: string, operator: any, value?: any): this;
    havingRaw(sql: string, bindings?: any[], conj?: string): this;
    orHavingRaw(sql: string, bindings?: any[]): this;
    orderBy(column: string, direction?: string): this;
    orderByRaw(sql: string): this;
    limit(n: number): this;
    offset(n: number): this;
    bindings(): any[];
    toSql(): string;
    /**
     * Exec update clause
     * @param map
     */
    update(map: Map): Promise<number>;
    insert(map: Map | Map[]): Promise<number>;
    insertGetId(map: Map, sequence?: string): Promise<number>;
    delete(): Promise<number>;
    truncate(): Promise<number>;
    /**
     * Query select clause and fetch the first one
     * @param columns
     */
    first(columns?: [string | Expression]): Promise<Map | null>;
    /**
     * Query select clause and fetch records
     * @param columns
     */
    get(columns?: [string | Expression]): Promise<Map[]>;
    /**
     * Fetch the column's value of first record
     * @param column
     */
    value(column: string): Promise<any | null>;
    /**
     * Fetch the column's values of records
     * @param column
     * @param key
     */
    pluck(column: string, key?: string): Promise<any[] | Map>;
    lastSql(): string;
    count(column?: string): Promise<number>;
    sum(column: string): Promise<number>;
    max(column: string): Promise<number>;
    min(column: string): Promise<number>;
    avg(column: string): Promise<number>;
    protected affectedRows(res: any): number;
    protected aggregate(func: string, column: string): Promise<number>;
    protected setAggregate(func: string, column: string): this;
    protected grammar(): Grammar;
    protected runSelect(): Promise<Map[]>;
    protected toRecords(res: any): Map[];
    protected pushJoin(type: string, table: string, query?: Builder): this;
    protected pushHavingBasic(conj: string, column: string, operator: string, value: any): this;
    protected pushHavingRaw(conj: string, sql: string, bindings?: any[]): this;
    protected pushWhereBasic(conj: string, column: string, operator: string, value: any): this;
    protected pushWhereColumn(conj: string, first: string, operator: string, second: string): this;
    protected pushWhereIn(conj: string, column: string, values: any[], not: boolean): this;
    protected pushWhereNull(conj: string, column: string, not: boolean): this;
    protected pushWhereNested(conj: string, query: Builder): this;
    protected pushWhereRaw(conj: string, sql: string, bindings?: any[]): this;
    protected addBindingsHaving(value: any | any[]): this;
    protected addBindingsUpdate(values: any[]): this;
    /**
     * insert values
     * @param values
     */
    protected addBindingsValues(values: any[]): this;
    protected addBindingsWhere(value: any | any[]): this;
    protected addBindingsJoin(value: any | any[]): this;
    protected addBindingsSelect(values: any[]): this;
    protected mergeBindings(): string[];
    protected clearBindings(): this;
    protected invalidOperator(operator: string): boolean;
}
