"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var Builder_1 = require("../Builder");
var Expression_1 = require("../Expression");
var helpers_1 = require("../../helpers");
var Grammar = /** @class */ (function () {
    function Grammar() {
        this.operators = [];
    }
    Grammar.prototype.getOperators = function () {
        return this.operators;
    };
    /**
     * Build the select statement sql
     * @param query
     */
    Grammar.prototype.compileSelect = function (query) {
        if (undefined === query._columns || 0 === query._columns.length) {
            query._columns = ["*"];
        }
        var sqlArr = [
            this.compileAggregate(query),
            this.compileColumn(query),
            this.compileFrom(query),
            this.compileJoins(query),
            this.compileWheres(query),
            this.compileGroups(query),
            this.compileHavings(query),
            this.compileOrders(query),
            this.compileLimit(query),
            this.compileOffset(query),
        ];
        return sqlArr.filter(function (sql) {
            return "" !== sql;
        }).join(" ");
    };
    Grammar.prototype.compileUpdate = function (query, map) {
        var _this = this;
        var table = this.wrap(query._from);
        // Each one of the columns in the update statements needs to be wrapped in the
        // keyword identifiers, also a place-holder needs to be created for each of
        // the values in the list of bindings so we can make the sets statements.
        var columnsArr = [];
        Object.keys(map).forEach(function (column) {
            columnsArr.push(_this.wrap(column) + " = " + _this.parameter());
        });
        var columns = columnsArr.join(", ");
        // If the query has any "join" clauses, we will setup the joins on the builder
        // and compile them so we can attach them to this update, as update queries
        // can get join statements to attach to other tables when they're needed.
        var joins = "";
        if (0 !== query._joins.length) {
            joins += " " + this.compileJoins(query); // contains an empty heading space
        }
        // Of course, update queries may also be constrained by where clauses so we'll
        // need to compile the where clauses and attach it to the query so only the
        // intended records are updated by the SQL statements we generate to run.
        var wheres = this.compileWheres(query);
        return "update " + table + joins + " set " + columns + " " + wheres;
    };
    Grammar.prototype.compileInsert = function (query, maps) {
        var _this = this;
        var columnsArr = [];
        Object.keys(maps[0]).forEach(function (column) {
            columnsArr.push(column);
        });
        var columns = columnsArr.join(", ");
        var valuesArr = [];
        maps.forEach(function (map) {
            valuesArr.push("(" + _this.parameterize(helpers_1.objectHelper.values(map)) + ")");
        });
        var values = valuesArr.join(", ");
        return "insert into " + this.wrap(query._from) + " (" + columns + ") values " + values;
    };
    Grammar.prototype.compileInsertGetId = function (query, map, sequence) {
        return this.compileInsert(query, [map]);
    };
    Grammar.prototype.compileDelete = function (query) {
        var select = this.compileSelect(query);
        return "delete " + select.substr(select.indexOf("from"));
    };
    Grammar.prototype.compileTruncate = function (query) {
        return "truncate table " + this.wrap(query._from);
    };
    /**
     * Convert an array of column names into a delimited string
     * @param columns
     */
    Grammar.prototype.columnize = function (columns) {
        var _this = this;
        var wrappedColumns = [];
        columns.forEach(function (col) {
            if ("*" === col) {
                wrappedColumns.push(col);
                return;
            }
            wrappedColumns.push(_this.wrap(col));
        });
        return wrappedColumns.join(", ");
    };
    /**
     * Wrap a value in keyword identifiers
     * @param value
     */
    Grammar.prototype.wrap = function (value) {
        if (value instanceof Expression_1.Expression) {
            return value.value();
        }
        var aliasedSegments = [];
        if (-1 !== value.toLowerCase().indexOf("as")) {
            value = this.wrapAliasedValue(value);
            aliasedSegments = value.split(/\s+as\s+/i);
        }
        var dotSegments = [];
        if (0 !== aliasedSegments.length) {
            dotSegments = aliasedSegments[0].split(".");
        }
        else {
            dotSegments = value.split(".");
        }
        value = this.wrapSingleValue(dotSegments[0]);
        if (dotSegments[1]) {
            value += "." + this.wrapSingleValue(dotSegments[1]);
        }
        if (aliasedSegments[1]) {
            value += " as " + aliasedSegments[1];
        }
        return value;
    };
    /**
     * Wrap a value that has an alias
     * @param value
     */
    Grammar.prototype.wrapAliasedValue = function (value) {
        var aliasedSegments = value.split(/\s+as\s+/i);
        value = aliasedSegments[0] + " as " + this.wrapSingleValue(aliasedSegments[1]);
        return value;
    };
    /**
     * Wrap a single string in keyword identifiers
     * @param value
     */
    Grammar.prototype.wrapSingleValue = function (value) {
        return value;
    };
    /**
     * Get the appropriate query parameter place-holder
     */
    Grammar.prototype.parameter = function (value) {
        if (value instanceof Expression_1.Expression) {
            return value.value();
        }
        return "?";
    };
    /**
     * Create query parameter place-holders for an array
     * @param params
     */
    Grammar.prototype.parameterize = function (params) {
        var _this = this;
        var placeHolders = [];
        params.forEach(function (param) {
            placeHolders.push(_this.parameter(param));
        });
        return placeHolders.join(", ");
    };
    /**
     * Compile an aggregated select clause
     * @param query
     */
    Grammar.prototype.compileAggregate = function (query) {
        if (undefined === query._aggregate) {
            return "";
        }
        var column = this.columnize([query._aggregate.Column]);
        if (query._distinct && column !== "*") {
            column += "distinct " + column;
        }
        return "select " + query._aggregate.Func + "(" + column + ") as aggregate";
    };
    /**
     * Compile the "select *" portion of the query
     * @param query
     */
    Grammar.prototype.compileColumn = function (query) {
        if (undefined !== query._aggregate) {
            return "";
        }
        var select = query._distinct ? "select distinct" : "select";
        return select + " " + this.columnize(query._columns);
    };
    /**
     * Compile the "from" portion of the query
     * @param query
     */
    Grammar.prototype.compileFrom = function (query) {
        return "from " + this.wrap(query._from);
    };
    /**
     * Compile the "join" portions of the query
     * @param query
     */
    Grammar.prototype.compileJoins = function (query) {
        if (undefined === query._joins || 0 === query._joins.length) {
            return "";
        }
        var join = this.compileJoinsToArray(query).join(" ");
        return join;
    };
    Grammar.prototype.compileJoinsToArray = function (query) {
        var _this = this;
        var joinsArr = [];
        query._joins.forEach(function (join) {
            var joinStr = join.Type + " join " + _this.wrap(join.Table);
            if (undefined !== join.Query) {
                joinStr += " on " + _this.compileWheres(join.Query).substr("where ".length);
            }
            joinsArr.push(joinStr);
        });
        return joinsArr;
    };
    /**
     * Compile the "where" portions of the query
     * @param query
     */
    Grammar.prototype.compileWheres = function (query) {
        if (undefined === query._wheres || 0 === query._wheres.length) {
            return "";
        }
        // If we actually have some where clauses, we will strip off the first boolean operator
        var wheresArr = this.compileWheresToArray(query);
        if (0 !== wheresArr.length) {
            return "where " + this.concatenateWhereClauses(wheresArr);
        }
        return "";
    };
    /**
     * Get an array of all the where clauses for the query
     * @param query
     */
    Grammar.prototype.compileWheresToArray = function (query) {
        var _this = this;
        var wheresArr = [];
        query._wheres.forEach(function (where) {
            // Add leading conjunction first
            var whereStr = where.Conjunction + " ";
            switch (where.Type) {
                case Builder_1.WhereType.Basic: {
                    whereStr += _this.whereBasic(where);
                    break;
                }
                case Builder_1.WhereType.Null: {
                    whereStr += _this.whereNull(where);
                    break;
                }
                case Builder_1.WhereType.In: {
                    whereStr += _this.whereIn(where);
                    break;
                }
                case Builder_1.WhereType.Nested: {
                    whereStr += _this.whereNested(where);
                    break;
                }
                case Builder_1.WhereType.Column: {
                    whereStr += _this.whereColumn(where);
                    break;
                }
                case Builder_1.WhereType.Raw: {
                    whereStr += _this.whereRaw(where);
                    break;
                }
            }
            wheresArr.push(whereStr);
        });
        return wheresArr;
    };
    /**
     * Format the where clause statements into one string
     * @param wheresArr
     */
    Grammar.prototype.concatenateWhereClauses = function (wheresArr) {
        return wheresArr.join(" ").replace(/and |or /i, "");
    };
    Grammar.prototype.whereBasic = function (where) {
        return this.wrap(where.Column) + " " + where.Operator + " " + this.parameter(where.Value);
    };
    Grammar.prototype.whereColumn = function (where) {
        return this.wrap(where.Column) + " " + where.Operator + " " + this.wrap(where.Value);
    };
    Grammar.prototype.whereRaw = function (where) {
        return where.Sql;
    };
    Grammar.prototype.whereNull = function (where) {
        if (where.Not) {
            return this.whereNotNull(where);
        }
        return this.wrap(where.Column) + " is null";
    };
    Grammar.prototype.whereNotNull = function (where) {
        return this.wrap(where.Column) + " is not null";
    };
    Grammar.prototype.whereIn = function (where) {
        if (where.Not) {
            return this.whereNotIn(where);
        }
        if (undefined === where.Values || 0 === where.Values.length) {
            return "0 = 1";
        }
        return this.wrap(where.Column) + " in (" + this.parameterize(where.Values) + ")";
    };
    Grammar.prototype.whereNotIn = function (where) {
        if (undefined === where.Values || 0 === where.Values.length) {
            return "1 = 1";
        }
        return this.wrap(where.Column) + " not in (" + this.parameterize(where.Values) + ")";
    };
    Grammar.prototype.whereNested = function (where) {
        // Remove the leading `where `
        return "(" + this.compileWheres(where.Query).substr("where ".length) + ")";
    };
    /**
     * Compile the "group by" portions of the query
     * @param query
     */
    Grammar.prototype.compileGroups = function (query) {
        if (undefined === query._groups || 0 === query._groups.length) {
            return "";
        }
        return "group by " + this.compileGroupsToArray(query._groups).join(", ");
    };
    /**
     * Compile the query groups to an array
     * @param groups
     */
    Grammar.prototype.compileGroupsToArray = function (groups) {
        var _this = this;
        var groupsArr = [];
        groups.forEach(function (group) {
            if (typeof group.Sql === "string" && 0 !== group.Sql.length) {
                groupsArr.push(group.Sql);
                return;
            }
            groupsArr.push(_this.wrap(group.Column));
        });
        return groupsArr;
    };
    /**
     * Compile the "having" portions of the query
     * @param query
     */
    Grammar.prototype.compileHavings = function (query) {
        if (0 === query._havings.length) {
            return "";
        }
        var havingsArr = this.compileHavingsToArray(query._havings);
        if (0 === havingsArr.length) {
            return "";
        }
        return "having " + this.concatenateHavingClause(havingsArr);
    };
    Grammar.prototype.compileHavingsToArray = function (havings) {
        var _this = this;
        var havingsArr = [];
        havings.forEach(function (having) {
            var havingStr = having.Conjunction + " ";
            if (typeof having.Sql === "string" && 0 !== having.Sql.length) {
                havingStr += having.Sql;
                havingsArr.push(havingStr);
                return;
            }
            havingStr += _this.wrap(having.Column) + " " + having.Operator + " " + _this.parameter();
            havingsArr.push(havingStr);
        });
        return havingsArr;
    };
    Grammar.prototype.concatenateHavingClause = function (havingsArr) {
        return havingsArr.join(" ").replace(/and |or /i, "");
    };
    /**
     * Compile the "order by" portions of the query
     * @param query
     */
    Grammar.prototype.compileOrders = function (query) {
        if (undefined === query._orders || 0 === query._orders.length) {
            return "";
        }
        return "order by " + this.compileOrdersToArray(query._orders).join(", ");
    };
    /**
     * Compile the query orders to an array
     * @param orders
     */
    Grammar.prototype.compileOrdersToArray = function (orders) {
        var _this = this;
        var orderArr = [];
        orders.forEach(function (order) {
            if (typeof order.Sql === "string" && 0 !== order.Sql.length) {
                orderArr.push(order.Sql);
                return;
            }
            orderArr.push(_this.wrap(order.Column) + " " + order.Direction);
        });
        return orderArr;
    };
    /**
     * Compile the "limit" portions of the query
     * @param query
     */
    Grammar.prototype.compileLimit = function (query) {
        if (undefined === query._limit) {
            return "";
        }
        return "limit " + query._limit;
    };
    /**
     * Compile the "offset" portions of the query
     * @param query
     */
    Grammar.prototype.compileOffset = function (query) {
        if (undefined === query._offset) {
            return "";
        }
        return "offset " + query._offset;
    };
    return Grammar;
}());
exports.Grammar = Grammar;
//# sourceMappingURL=Grammar.js.map